#ifndef __ManualIntersector_h__
#define __ManualIntersector_h__

#include "File.h"
#include "Queue.h"
#include "Stringf.h"
#include "ThreadStateGraph.h"

class ManualIntersector
{
public:
    void intersect( ThreadStateGraph &threadOne, ThreadStateGraph &threadTwo, ThreadStateGraph &output, char separationChar )
    {
        storeSeparationChar( separationChar );
        storeGraphs( threadOne, threadTwo, output );
        initializeOutputGraph();

        while ( haveStatesToCheck() )
        {
            checkAnotherState();
        }
    }

private:
    ThreadStateGraph *_threadOne;
    ThreadStateGraph *_threadTwo;
    ThreadStateGraph *_output;

    Map<String, String> _projectToFirst;
    Map<String, String> _projectToSecond;

    Map<String, char> _statesAlreadyChecked;
    Queue<String> _toCheck;

    char _separationChar;

    void storeSeparationChar( char separationChar )
    {
        _separationChar = separationChar;
    }

    bool haveStatesToCheck() const
    {
        return !_toCheck.empty();
    }

    void storeGraphs( ThreadStateGraph &threadOne, ThreadStateGraph &threadTwo, ThreadStateGraph &output )
    {
        _threadOne = &threadOne;
        _threadTwo = &threadTwo;
        _output = &output;
    }

    void initializeOutputGraph()
    {
        _output->_events = _threadOne->_events;

        _output->_atomicPropositions = _threadOne->_atomicPropositions;
        _output->_atomicPropositions.add( _threadTwo->_atomicPropositions );

        _output->_threadName = _threadOne->_threadName + " " + _threadTwo->_threadName;

        createInitialState();
    }

    void createInitialState()
    {
        String newInitial = addCombinedState( _threadOne->_initialState, _threadTwo->_initialState );
        //TS_TRACE( Stringf( "New initial state: %s", newInitial.ascii() ).ascii() );

        _output->_initialState = newInitial;
    }

    String addCombinedState( const String &labelOne, const String &labelTwo )
    {
        String newLabel = fuseLabels( labelOne, labelTwo );

        ThreadState newState;
        newState._label = newLabel;

        fuseStates( labelOne, labelTwo, newState );

        _projectToFirst[newLabel] = labelOne;
        _projectToSecond[newLabel] = labelTwo;

        _output->_graph[newLabel] = newState;

        if ( !_statesAlreadyChecked.exists( newLabel ) )
        {
            _toCheck.push( newLabel );
            _statesAlreadyChecked[newLabel] = 0;
        }

        return newLabel;
    }

    String fuseLabels( const String &labelOne, const String &labelTwo ) const
    {
        return Stringf( "%s%c%s", labelOne.ascii(), _separationChar, labelTwo.ascii() );
    }

    String projectToFirst( String &fusedLabel )
    {
        return _projectToFirst[fusedLabel];
    }

    String projectToSecond( String &fusedLabel )
    {
        return _projectToSecond[fusedLabel];
    }

    void fuseStates( const String &first, const String &second, ThreadState &fused )
    {
        ThreadState &stateOne( _threadOne->_graph[first] );
        ThreadState &stateTwo( _threadTwo->_graph[second] );

        fused._requestedEvents = stateOne._requestedEvents + stateTwo._requestedEvents;
        fused._blockedEvents = stateOne._blockedEvents + stateTwo._blockedEvents;

        fused._requestedEvents.uniqueElements();
        fused._blockedEvents.uniqueElements();

        for ( unsigned i = 0; i < fused._blockedEvents.size(); ++i )
        {
            Event blockedEvent = fused._blockedEvents[i];

            if ( fused._requestedEvents.exists( blockedEvent ) )
            {
                fused._requestedEvents.eraseByValue( fused._blockedEvents[i] );
            }
        }

        for ( unsigned i = 0; i < stateOne._atomicPropositions.size(); ++i )
        {
            fused._atomicPropositions.add( stateOne._atomicPropositions.at( i ) );
        }

        for ( unsigned i = 0; i < stateTwo._atomicPropositions.size(); ++i )
        {
            fused._atomicPropositions.add( stateTwo._atomicPropositions.at( i ) );
        }
    }

    Vector<Event> nonBlockedEventsOfState( String &label )
    {
        Vector<Event> results = _output->_events;

        for ( unsigned i = 0; i < _output->_graph[label]._blockedEvents.size(); ++i )
        {
            results.eraseByValue( _output->_graph[label]._blockedEvents[i] );
        }

        return results;
    }

    void checkAnotherState()
    {
        String label = _toCheck.peak();
        _toCheck.pop();

        Vector<Event> nonBlockedEvents = nonBlockedEventsOfState( label );

        for ( unsigned i = 0; i < nonBlockedEvents.size(); ++i )
        {
            addEdgesForEvent( label, nonBlockedEvents[i] );
        }
    }

    void addEdgesForEvent( String &label, Event &event )
    {
        ThreadState &first( _threadOne->_graph[projectToFirst( label )] );
        ThreadState &second( _threadTwo->_graph[projectToSecond( label )] );

        Vector<String> successorsOfFirst = first.labelsOfSuccessors( event );
        Vector<String> successorsOfSecond = second.labelsOfSuccessors( event );

        for ( unsigned i = 0; i < successorsOfFirst.size(); ++i )
        {
            for ( unsigned j = 0; j < successorsOfSecond.size(); ++j )
            {
                String successor = fuseLabels( successorsOfFirst[i], successorsOfSecond[j] );

                if ( !_output->_graph.exists( successor ) )
                {
                    addCombinedState( successorsOfFirst[i], successorsOfSecond[j] );
                }

                addSingleEdge( label, event, successor );
            }
        }
    }

    void addSingleEdge( String &state, Event &event, String &successor )
    {
        ThreadState &source( _output->_graph[state] );
        ThreadState &target( _output->_graph[successor] );

        Transition transition( event, successor );
        source._successors.append( transition );

        Transition backwardTransition( event, state );
        target._predecessors.append( backwardTransition );
    }
};

#endif // __ManualIntersector_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
