#ifndef __SafetyModelChecker_h__
#define __SafetyModelChecker_h__

static const char BAD_ATOMIC_PROPOSITION[] = "BAD_STATE";

class SafetyModelChecker
{
public:
    static void checkDeadlocks( ThreadStateGraph &stateGraph )
    {
        Map<String, ThreadState>::iterator it;

        for ( it = stateGraph._graph.begin(); it != stateGraph._graph.end(); ++it )
        {
            TS_ASSERT( !it->second._successors.empty() );
        }

        TS_TRACE( "No deadlocks found!" );
    }

    static bool checkSafetyProperty( ThreadStateGraph &programGraph, NFA &property, NFA::Run &counterExample, NFA &productAutomaton )
    {
        return SafetyModelChecker( programGraph, property, productAutomaton ).run( counterExample );
    }

private:
    ThreadStateGraph *_programGraph;
    NFA *_property;
    NFA *_productAutomaton;

    Map<String, String> _projectToProgram;
    Map<String, String> _projectToProperty;

    SafetyModelChecker( ThreadStateGraph &programGraph, NFA &property, NFA &productAutomaton )
    {
        storeInput( programGraph, property, productAutomaton );
    }

    bool run( NFA::Run &counterExample )
    {
        TS_TRACE( "Safety model checker running. Composing state graph and property." );

        if ( productAutomatonEmpty() )
        {
            constructProductAutomaton();
        }
        else
        {
            _productAutomaton->eliminateNonReachables();
        }

        TS_TRACE( Stringf( "Product automaton constructed. States: %u. Edges: %u.", _productAutomaton->numberOfStates(), _productAutomaton->edgeCount() ).ascii() );
        // TS_TRACE( "Dumping product automaton..." );
        // _productAutomaton->dump();

        if ( _productAutomaton->emptyLanguage() )
        {
            return true;
        }

        counterExample = _productAutomaton->findRunInLanguage();
        return false;
    }

    bool productAutomatonEmpty() const
    {
        return _productAutomaton->numberOfStates() == 0;
    }

    void storeInput( ThreadStateGraph &programGraph, NFA &property, NFA &productAutomaton )
    {
        _programGraph = &programGraph;
        _property = &property;
        _productAutomaton = &productAutomaton;

        //        _productAutomaton->clear();
    }

    void constructProductAutomaton()
    {
        initAutomaton();
        constructStates();
        constructEdges();
        eliminateNonReachables();
    }

    void initAutomaton()
    {
        for ( unsigned i = 0; i < _programGraph->_events.size(); ++i )
        {
            _productAutomaton->addLetter( programEventToProductEvent( _programGraph->_events[i] ) );
        }

        String initialStateAp = _programGraph->_graph[_programGraph->_initialState]._atomicPropositions.toString();
        String propertySuccessor = _property->getSuccessor( _property->getInitialState(), initialStateAp );

        _productAutomaton->setInitialState( fuseStates( _programGraph->_initialState, propertySuccessor ) );
    }

    static String programEventToProductEvent( const Event &event )
    {
        return Stringf( "%u", event.type() );
    }

    void constructStates()
    {
        Map<String, ThreadState>::iterator it;

        for ( it = _programGraph->_graph.begin(); it != _programGraph->_graph.end(); ++it )
        {
            for ( unsigned i = 0; i < _property->numberOfStates(); ++i )
            {
                String programState = it->first;
                String propertyState = _property->getState( i );
                String fusedState = fuseStates( programState, propertyState );

                _projectToProgram[fusedState] = programState;
                _projectToProperty[fusedState] = propertyState;

                _productAutomaton->addState( fusedState );

                if ( _property->acceptingState( propertyState ) )
                {
                    _productAutomaton->markAsAccepting( fusedState );
                }
            }
        }
    }

    static String fuseStates( const String &programState, const String &propertyState )
    {
        return Stringf( "%s/%s", programState.ascii(), propertyState.ascii() );
    }

    void constructEdges()
    {
        for ( unsigned i = 0; i < _productAutomaton->numberOfStates(); ++i )
        {
            constructEdgesForState( _productAutomaton->getState( i ) );
        }
    }

    void constructEdgesForState( const String &state )
    {
        String programState = _projectToProgram[state];
        String propertyState = _projectToProperty[state];

        for ( unsigned i = 0; i < _programGraph->_graph[programState]._successors.size(); ++i )
        {
            Transition programTransition = _programGraph->_graph[programState]._successors[i];
            String programSuccessor = programTransition._state;
            String successorPropositions = _programGraph->_graph[programSuccessor]._atomicPropositions.toString();

            // printf( "Working on transition: %s --> %s (event %u)\n", programState.ascii(), programSuccessor.ascii(), programTransition._event.type() );
            if ( _property->existsSuccessor( propertyState, successorPropositions ) )
            {
                String propertySuccessor = _property->getSuccessor( propertyState, successorPropositions );
                String productSuccessor = fuseStates( programSuccessor, propertySuccessor );

                _productAutomaton->addEdge( NFA::Edge( state, programEventToProductEvent( programTransition._event ), productSuccessor ) );
            }
            else
            {
                throw Error( Error::NO_SUCCESSOR_IN_NFA );
            }
        }
    }

    void eliminateNonReachables()
    {
        _productAutomaton->eliminateNonReachables();
    }
};

#endif // __SafetyModelChecker_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
