#ifndef __TcpFinHandler_h__
#define __TcpFinHandler_h__

#include "BThread.h"
#include "IEvent.h"
#include "ServerConfiguration.h"

class TcpFinHandler : public BThread
{
public:
    TcpFinHandler()
    {
    }

    void entryPoint()
    {
        while ( true )
        {
            waitForTcpFinRequest();
            storeFinRequest();
            getFreshSequenceNumber();
            replyWithFinAck();
            doneHandlingIpPacket();
        }
    }

private:
    unsigned _sequenceNumber;
    TcpSegment _finRequest;

    void storeFinRequest()
    {
        _finRequest.unserialize( lastEvent().payload() );
    }

    void waitForTcpFinRequest()
    {
        clearEventVectors();

        _watched.append( Event( IEvent::TCP_FIN_REQUEST ) );
        _blocked = allRelevantEventsExcept( IEvent::TCP_FIN_REQUEST );
        bSync( _requested, _watched, _blocked, "Wait" );
    }

    void replyWithFinAck()
    {
        TcpSegment finAck;
        prepareAckSegment( finAck );
        sendFinAck( finAck );
    }

    void prepareAckSegment( TcpSegment &finAck )
    {
        finAck.header().setSourcePort( _finRequest.header().getDestinationPort() );
        finAck.header().setDestinationPort( _finRequest.header().getSourcePort() );

        finAck.header().setAckNumber( _finRequest.header().getSequenceNumber() + 1 );
        finAck.header().setSequenceNumber( _sequenceNumber );

        finAck.header().setAck();
        finAck.header().setFin();

        finAck.header().setWindowSize( ServerConfiguration::TCP_WINDOW_SIZE );
    }

    void sendFinAck( TcpSegment &finAck )
    {
        HeapData serialized;
        finAck.serialize( serialized );

        clearEventVectors();
        _requested.append( Event( IEvent::TCP_SEND_DATA, serialized ) );
        _blocked = allRelevantEventsExcept( IEvent::TCP_SEND_DATA );

        AtomicPropositions sendFinAck;
        sendFinAck.add( "SendFinAck" );

        bSync( _requested, _watched, _blocked, sendFinAck, "FinAck" );
    }

    void doneHandlingIpPacket()
    {
        clearEventVectors();
        _requested.append( Event( IEvent::TCP_DONE_HANDLING_DATA ) );
        _blocked = allRelevantEventsExcept( IEvent::TCP_DONE_HANDLING_DATA );
        bSync( _requested, _watched, _blocked, "Done" );
    }

    void getFreshSequenceNumber()
    {
        requestSequenceNumber();
        readSequenceNumber();
    }

    void requestSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
            FIN_REQUEST_COUNTS_AS_ONE_BYTE = 1,
        };

        clearEventVectors();

        SequenceNumberPayload sequenceNumber( DONT_CARE, FIN_REQUEST_COUNTS_AS_ONE_BYTE );

        _requested.append( Event( IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER, ConstSimpleData( &sequenceNumber, sizeof(sequenceNumber) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        bSync( _requested, _watched, _blocked, "RequestSeq" );
    }

    void readSequenceNumber()
    {
        clearEventVectors();

        _watched.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );
        _blocked = allRelevantEventsExcept( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );
        bSync( _requested, _watched, _blocked, "ReadSeq" );

        _sequenceNumber = SequenceNumberPayload( lastEvent().payload() ).getSequenceNumber();
    }

    Vector<Event> allRelevantEventsExcept( Event event )
    {
        Vector<Event> result;

        if ( event != IEvent::TCP_FIN_REQUEST ) result.append( IEvent::TCP_FIN_REQUEST );
        if ( event != IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER ) result.append( IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        if ( event != IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER ) result.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );

        return result;
    }
};

#endif // __TcpFinHandler_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
