#ifndef __TcpSynHandler_h__
#define __TcpSynHandler_h__

#include "BThread.h"
#include "IEvent.h"
#include "ServerConfiguration.h"
#include "TargetPortPayload.h"

class TcpSynHandler : public BThread
{
public:
    TcpSynHandler()
    {
    }

    void entryPoint()
    {
        while ( true )
        {
            waitForTcpSynRequest();
            storeSynRequest();
            getFreshSequenceNumber();
            replyWithSynAck();
            broadcastTargetPort();
            doneHandlingIpPacket();
        }
    }

private:
    unsigned _sequenceNumber;
    TcpSegment _synRequest;

    void storeSynRequest()
    {
        _synRequest.unserialize( lastEvent().payload() );
    }

    void waitForTcpSynRequest()
    {
        clearEventVectors();

        _watched.append( IEvent::TCP_SYN_REQUEST );
        _blocked = allRelevantEventsExcept( IEvent::TCP_SYN_REQUEST );
        bSync( _requested, _watched, _blocked, "Wait" );
    }

    void replyWithSynAck()
    {
        TcpSegment synAck;
        prepareAckSegment( synAck );
        sendSynAck( synAck );
    }

    void prepareAckSegment( TcpSegment &synAck )
    {
        synAck.header().setSourcePort( _synRequest.header().getDestinationPort() );
        synAck.header().setDestinationPort( _synRequest.header().getSourcePort() );

        synAck.header().setAckNumber( _synRequest.header().getSequenceNumber() + 1 );
        synAck.header().setSequenceNumber( _sequenceNumber );

        synAck.header().setAck();
        synAck.header().setSyn();

        synAck.header().setWindowSize( ServerConfiguration::TCP_WINDOW_SIZE );
    }

    void sendSynAck( TcpSegment &synAck )
    {
        HeapData serialized;
        synAck.serialize( serialized );

        clearEventVectors();

        _requested.append( Event( IEvent::TCP_SEND_DATA, serialized ) );
        _blocked = allRelevantEventsExcept( IEvent::TCP_SEND_DATA );

        AtomicPropositions sendSynAck;
        sendSynAck.add( "SendSynAck" );

        bSync( _requested, _watched, _blocked, sendSynAck, "SynAck" );
    }

    void doneHandlingIpPacket()
    {
        clearEventVectors();
        _requested.append( IEvent::TCP_DONE_HANDLING_DATA );
        _blocked = allRelevantEventsExcept( IEvent::TCP_DONE_HANDLING_DATA );
        bSync( _requested, _watched, _blocked, "Done" );
    }

    void getFreshSequenceNumber()
    {
        requestSequenceNumber();
        readSequenceNumber();
    }

    void requestSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
            SYN_REQUEST_COUNTS_AS_ONE_BYTE = 1,
        };

        clearEventVectors();

        SequenceNumberPayload sequenceNumber( DONT_CARE, SYN_REQUEST_COUNTS_AS_ONE_BYTE );

        _requested.append( Event( IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER, ConstSimpleData( &sequenceNumber, sizeof(sequenceNumber) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        bSync( _requested, _watched, _blocked, "RequestSeq" );
    }

    void broadcastTargetPort()
    {
        clearEventVectors();

        TargetPortPayload targetPortPayload( _synRequest.header().getSourcePort() );

        _requested.append( Event( IEvent::TCP_ANNOUNCE_TARGET_PORT, ConstSimpleData( &targetPortPayload, sizeof(targetPortPayload) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::TCP_ANNOUNCE_TARGET_PORT );
        bSync( _requested, _watched, _blocked, "BroadcastPort" );
    }

    void readSequenceNumber()
    {
        clearEventVectors();

        _watched.append( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );
        _blocked = allRelevantEventsExcept( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );
        bSync( _requested, _watched, _blocked, "ReadSeq" );

        _sequenceNumber = SequenceNumberPayload( lastEvent().payload() ).getSequenceNumber();
    }

    Vector<Event> allRelevantEventsExcept( Event event )
    {
        Vector<Event> result;

        if ( event != IEvent::TCP_SYN_REQUEST ) result.append( IEvent::TCP_SYN_REQUEST );
        if ( event != IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER ) result.append( IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        if ( event != IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER ) result.append( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );
        if ( event != IEvent::TCP_ANNOUNCE_TARGET_PORT ) result.append( IEvent::TCP_ANNOUNCE_TARGET_PORT );

        return result;
    }
};

#endif // __TcpSynHandler_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
